"""
Testing problem180meridian.split180_multilinestring()
"""

from ..problem180meridian import split180_multilinestring


# Testing integer coordinates without crossing 180th meridian
def test_split180_int_nochange():

    test_coordinates = [((0, 0), (0, 10), (10, 10), (10, 0)), ((100, 100), (100, 110), (110, 110), (110, 100))]
    output = 'MULTILINESTRING ((0 0,0 10,10 10,10 0),(100 100,100 110,110 110,110 100))'

    test_output = split180_multilinestring(test_coordinates).ExportToWkt()
    print(test_output)
    assert output == test_output


# Testing float coordinates without crossing 180th meridian
def test_split180_float_nochange():

    test_coordinates = [((160.27414, 66.38205), (165.87357, 65.61501), (160.50426, 58.94171), (166.21875, 62.08660))]
    output = 'MULTILINESTRING ((160.27414 66.38205,165.87357 65.61501,160.50426 58.94171,166.21875 62.0866))'

    test_output = split180_multilinestring(test_coordinates).ExportToWkt()
    print(test_output)
    assert output == test_output



"""
Testing problem180meridian.points_from_wkt()
"""

from osgeo import ogr

from ..problem180meridian.get_coordinates import points_from_geometry


def check_parsing(feature):
    geometry = feature.GetGeometryRef()
    points, geometry_type = points_from_geometry(geometry)
    if points is not None:
        print(points)


test_ds = ogr.Open('./test/test_data/test_geometry.geojson')
test_lr = test_ds.GetLayer()


# Testing POINT geometry for cross180 - None is expected

def test_point():
    check_parsing(test_lr.GetFeature(1)) is None


# Testing LINESTRING geometry for cross180 - None is expected

def test_linestring():
    check_parsing(test_lr.GetFeature(0)) == [[[(24.5415438402914, 13.5588583164334), (25.5100146734532, 4.19366085458475),(32.6431265127727, 6.13284981165079)]]]


# Testing POLYGON geometry for cross180 - None is expected

def test_polygon():
    check_parsing(test_lr.GetFeature(2)) == [[[(42.2673084292096, 59.944926664255), (75.0211577247333, 62.6922826805411), (47.1942850696495, 67.2099881248319), (23.0620579829856, 63.2777691933206), (26.5176224655007, 66.1661342312123), (15.0717845363254, 58.7370412696996), (42.2673084292096, 59.9449266642551)]]]



import numpy as np

try:
    from osgeo import ogr
except ImportError:
    import ogr

from .polygonhierarchy import PolygonHierarchy
from .get_coordinates import points_from_geometry


__all__ = ['check180', 'cross180', 'split180_coordinates',
           'split180_multipolygon', 'split180_multilinestring',
           'split180_geometry']


# Returns True if value is positive or zero and False if it is negative
def array_not_negative(array):
    return np.sign(np.sign(array) + 1).astype(bool)


# Checks if any segment in a pair of coordinates crosses 180th meridian
# It's considered so if abs(x2 - x1) > abs(x2_1 - x1_1) where
# xi_1 = xi + (xi < 0) * 360
# If any crossing is found a check_array is returned where crossing pairs
# are marked as True
def check180(coordinates):

    longitudes = np.array([p[0] for p in coordinates])
    signs = array_not_negative(longitudes)

    if list(np.unique(signs)) == [0, 1]:
        distances = abs(longitudes[1:] - longitudes[:-1])
        longitudes_fixed = longitudes + ((signs == 0) * 360)
        distances_fixed = abs(longitudes_fixed[1:] - longitudes_fixed[:-1])
        # Need to add small value below to avoid Python calculation errors
        check_array = array_not_negative(distances_fixed - distances + 0.000001)
        return True, check_array
    else:
        return False, None


# Create coordinates for two points in a segment crossing 180th meridian
# Setting lon_buffer one can separate them from each other and
# the meridian itself
def cross180(coord1, coord2, lon_buffer=0):

    if coord1[0] > 0:
        assert coord2[0] < 0
        img_lon2 = coord2[0] + 360
        new_lon1 = 180 - abs(lon_buffer)
    else:
        assert coord2[0] >= 0
        img_lon2 = coord2[0] - 360
        new_lon1 = abs(lon_buffer) - 180

    if (img_lon2 - coord1[0]) == 0:
        angle_coefficient = 1
    else:
        angle_coefficient = (coord2[1] - coord1[1]) / (img_lon2 - coord1[0])

    new_lon2 = (-1) * new_lon1
    new_lat1 = coord1[1] + angle_coefficient * (new_lon1 - coord1[0])
    new_lat2 = coord2[1] + angle_coefficient * (new_lon2 - coord2[0])

    return (new_lon1, new_lat1), (- new_lon1, new_lat2)


# Split coordinates by the 180th meridian creating new coordinate chains
def split180_coordinates(coordinates, check, lon_buffer=0):

    assert len(coordinates) - 1 == len(check)

    chains = [[coordinates[0]]]

    for i in range(len(check)):

        if check[i]:
            chains[-1].append(coordinates[i+1])
        else:
            new_coord1, new_coord2 = cross180(
                *tuple(coordinates[i:i + 2]),
                lon_buffer=lon_buffer)
            chains[-1].append(new_coord1)
            chains.append([new_coord2, coordinates[i+1]])

    return chains


# Split linestring/multilinestring geometry by the 180th meridian
def split180_multilinestring(coordinates, lon_buffer=0):

    chains = []

    for poly in coordinates:

        cross, check = check180(coordinates[0])

        if cross:
            chain = split180_coordinates(poly, check, lon_buffer=lon_buffer)
            chains.extend(chain)
        else:
            chains.append(poly)

    wkt = 'MULTILINESTRING ((%s))' % '),('.join(
        [
            ','.join([f"{p[0]} {p[1]}" for p in chain]) for chain in chains
        ]
    )

    multilinestring = ogr.Geometry(wkt=wkt)

    return multilinestring


# Converts coordinate chains into a list of MultiPolygon geometry
def coordinate_chains_to_multipolygon(chains):

    # Chains from split180_coordinates are expected;
    # need to finish all the polygon chains to make correct wkt
    if len(chains) > 1:
        chains[0].extend(chains.pop())
        for chain in chains[1:]:
            chain.append(chain[0])

    wkt_list = [
        'MULTIPOLYGON (((%s)))' %
        ','.join(['%f %f' % (p[0], p[1]) for p in chain]) for chain in chains
    ]

    geom_list = [ogr.Geometry(wkt=wkt) for wkt in wkt_list]
    geoms = PolygonHierarchy(*geom_list)
    multipolygon = geoms.multipolygon()

    return multipolygon


# Split polygon/multipolygon geometry coordinates by the 180th meridian
def split180_multipolygon(coordinates, lon_buffer=0):

    geoms = PolygonHierarchy()

    for polygon in coordinates:
        for arring in polygon:

            cross, check = check180(arring)

            if cross:
                chains = split180_coordinates(arring, check, lon_buffer=lon_buffer)
                multipolygon = coordinate_chains_to_multipolygon(chains)
                geoms.join_geometry(multipolygon)
            else:
                point_str = ','.join(['%f %f' % (p[0], p[1]) for p in arring])
                wkt = f"MULTIPOLYGON ((({point_str})))"
                geoms.join_geometry(ogr.Geometry(wkt=wkt))

    return geoms.multipolygon()


# Split ogr.Geometry by the 180th meridian
def split180_geometry(geometry, lon_buffer=0):

    coordinates, geometry_type = points_from_geometry(geometry)

    if geometry_type is None:
        raise Exception('Geometry type not found')

    elif geometry_type in ('POINT', 'MULTIPOINT'):
        pass

    elif geometry_type in ('LINESTRING', 'MULTILINESTRING'):
        return split180_multilinestring(coordinates, lon_buffer=lon_buffer)

    elif geometry_type in ('POLYGON', 'MULTIPOLYGON'):
        return split180_multipolygon(coordinates, lon_buffer=lon_buffer)

    else:
        raise Warning(f'Unsupported geometry type {geometry_type}')

    return geometry



"""
Testing problem180meridian.points_from_wkt()
"""

from ..problem180meridian.get_coordinates import points_from_wkt


def check_parsing(wkt):
    points, geometry_type = points_from_wkt(wkt)
    if points is not None:
        print(points)


# Testing point wkt for cross180 - None is expected

def test_point():
    check_parsing('POINT (10 20)') is None

def test_multipoint():
    check_parsing('MULTIPOINT ((10 40), (40 30), (20 20), (30 10))') is None
    check_parsing('MULTIPOINT (10 40, 40 30, 20 20, 30 10)') is None


# Testing non processed wkt types for cross180 - None is expected

def test_circularstring():
    check_parsing('CIRCULARSTRING (1 5, 6 2, 7 3)') is None


# Testing empty wkt for cross180 - None is expected

def test_empty_polygon():
    check_parsing('POLYGON EMPTY') is None


# Testing linear wkt

def test_linestring():
    check_parsing('LINESTRING (10 20, 20 30)') == [[[(10.0, 20.0), (20.0, 30.0)]]]


# Testing polygon wkt

def test_polygon():
    check_parsing('MULTIPOLYGON (((40 40, 20 45, 45 30, 40 40)), ((20 35, 10 30, 10 10, 30 5, 45 20, 20 35), (30 20, 20 15, 20 25, 30 20)))') == [[[(40.0, 40.0), (20.0, 45.0), (45.0, 30.0), (40.0, 40.0)]], [[(20.0, 35.0), (10.0, 30.0), (10.0, 10.0), (30.0, 5.0), (45.0, 20.0), (20.0, 35.0)], [(30.0, 20.0), (20.0, 15.0), (20.0, 25.0), (30.0, 20.0)]]]



"""
Testing problem180meridian.split180_geometry()
"""

from osgeo import ogr

from ..problem180meridian import split180_geometry


def check_parsing(feature):
    geometry = feature.GetGeometryRef()
    new_geometry = split180_geometry(geometry)
    if new_geometry is not None:
        print(new_geometry)
    return new_geometry


test_ds = ogr.Open('./test/test_data/test_geometry.geojson')
test_lr = test_ds.GetLayer()


# Testing POINT geometry for cross180 - None is expected

def test_point():
    check_parsing(test_lr.GetFeature(1)) is None


# Testing LINESTRING geometry for cross180 - None is expected

def test_linestring():
    check_parsing(test_lr.GetFeature(0)).ExportToWkt() == 'MULTILINESTRING ((24.5415438402914 13.5588583164334, 25.5100146734532 4.19366085458475,32.6431265127727 6.13284981165079))'


# Testing POLYGON geometry for cross180 - None is expected

def test_polygon():
    check_parsing(test_lr.GetFeature(2)).ExportToWkt() == 'MULTIPOLYGON (((42.2673084292096 59.944926664255, 75.0211577247333 62.6922826805411, 47.1942850696495 67.2099881248319, 23.0620579829856 63.2777691933206, 26.5176224655007 66.1661342312123, 15.0717845363254 58.7370412696996, 42.2673084292096 59.9449266642551)))'



from osgeo import ogr


__all__ = ['PolygonHierarchy']


class GeometriesError(Exception):

    def __init__(self, *args, **kwargs):
        pass


# Arranges a set of polygons to find out which of them are within others
# If any of them is within others they are cut off while creating a new
# multipolygon
class PolygonHierarchy:

    def __init__(self, *geom_tuple, main_geom=None, level=0):

        assert (main_geom is None) or isinstance(main_geom, ogr.Geometry)

        self.main_geometry = main_geom
        self.level = level
        self.geometries = []

        for geometry in geom_tuple:
            self.join_geometry(geometry)

    def __str__(self):

        if self.main_geometry is None:
            main_wkt = 'No main geometry'
        else:
            main_wkt = self.main_geometry.ExportToWkt()

        tab = '\n\t' + self.level*'\t'

        return tab.join([main_wkt] + [str(h) for h in self.geometries])

    def check_main_geometry(self, geometry):

        if self.main_geometry is not None:

            if geometry.Within(self.main_geometry):
                # print('Geometry within')
                return 0

            elif self.main_geometry.Within(geometry):
                # print('Hierarchy within')
                return 2

            elif geometry.Intersects(self.main_geometry):
                raise GeometriesError('Unexpected intersection')

            else:
                # print('Out of main geometry')
                return 1

    def join_geometry(self, geometry):

        assert isinstance(geometry, ogr.Geometry)

        main_geometry_check = self.check_main_geometry(geometry)

        if main_geometry_check:
            return main_geometry_check

        new_geoms = PolygonHierarchy(main_geom=geometry, level=self.level + 1)
        separate = True

        for i in range(len(self.geometries)-1, -1, -1):

            self_geoms = self.geometries[i]
            separate = self.geometries[i].join_geometry(geometry)

            if separate == 0:
                # print('Joined hierarchy')
                return 0

            elif separate == 2:
                # Doesn't check if there are intersections within new_geoms
                # Need to fix if used for other polygons but raster footprints!
                new_geoms.geometries.append(self.geometries.pop(i))

            elif geometry.Intersects(self_geoms.main_geometry):
                raise GeometriesError('Unexpected intersection')

        if separate:
            self.geometries.append(new_geoms)
            return 0

        raise GeometriesError('Geometry not processed')

    def multipolygon(self):

        if self.main_geometry is not None:
            multipolygon = self.main_geometry
            for geoms in self.geometries:
                multipolygon = multipolygon.Difference(geoms.multipolygon())

        elif len(self.geometries) != 0:
            multipolygon = self.geometries[0].multipolygon()
            for geoms in self.geometries[1:]:
                multipolygon = multipolygon.Union(geoms.multipolygon())

        else:
            multipolygon = ogr.Geometry(6)

        return multipolygon



#!/usr/bin/env python

from io import open
from setuptools import setup

"""

:authors: sergsadkov
:license: Apache License, Version 2.0, see LICENSE file
:copyright: (c) 2023 sergsadkov
"""


version = '0.1.4'

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='problem180meridian',
    version=version,

    author='sergsadkov',
    author_email='sergsadkov@gmail.com',

    description=u'Fixes 180 degree crossing issue for vector geodata',

    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/sergsadkov/problem180meridian',
    download_url='https://github.com/sergsadkov/problem180meridian/archive/main.zip',

    license='Apache License, Version 2.0, see LICENSE file',

    packages=['problem180meridian'],
    install_requires=['numpy', 'GDAL'],

    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
    ]
)



# Functions to extract points from geometry
import re

from osgeo import ogr


__all__ = ['points_from_wkt', 'points_from_geometry', 'wkt_close_chain', 'wkt_brackets']


class WktParsingException(Exception):

    def __init__(self, *args, **kwargs):
        pass


def float_xy_coordinates(coordinate_string):
    coordinate_list = coordinate_string.strip().split(' ')[:2]
    x = float(coordinate_list[0])
    y = float(coordinate_list[1])
    return x, y


def parse_wkt_chain(wkt_coordinates_chain):
    coordinate_string_list = wkt_coordinates_chain.strip('() ').split(',')
    coordinate_list = [float_xy_coordinates(coordinate) for coordinate in coordinate_string_list]
    return coordinate_list


def parse_wkt_multichain(wkt_multichain, geom_type):

    if geom_type == 'LINESTRING':
        return [parse_wkt_chain(wkt_multichain)]

    elif geom_type == 'MULTILINESTRING':
        chains = wkt_multichain.strip('() ').split('),')
        return [parse_wkt_chain(chain) for chain in chains]

    elif geom_type == 'POLYGON':
        chains = wkt_multichain.strip('() ').split('),')
        return [[parse_wkt_chain(chain) for chain in chains]]

    elif geom_type == 'MULTIPOLYGON':
        chains2 = wkt_multichain.strip('() ').split(')),')
        return [[parse_wkt_chain(chain) for chain in chains.split('),')] for chains in chains2]

    else:
        raise WktParsingException('Wrong geometry type:', geom_type)


def points_from_wkt(wkt):

    geometry_type = re.search('^[A-Z]+', wkt).group()

    if geometry_type in ('POINT', 'MULTIPOINT'):
        print('No crossing for type:', geometry_type)
        return None, geometry_type

    elif geometry_type not in ('LINESTRING', 'MULTILINESTRING', 'POLYGON', 'MULTIPOLYGON'):
        print('Incorrect geometry type for crossing check:', geometry_type)
        return None, geometry_type

    elif re.search('EMPTY$', geometry_type):
        print('Geometry is empty:', wkt)
        return None, geometry_type

    wkt_multichain_search = re.search(r'\([()\-\d\., ]+\)', wkt)

    if wkt_multichain_search is None:
        print('Wkt coordinates not found:', wkt)
        return None, geometry_type

    else:
        coordinates = parse_wkt_multichain(wkt_multichain_search.group(), geometry_type)
        return coordinates, geometry_type


def wkt_close_chain(geometry_type):
    if geometry_type in ('LINESTRING', 'MULTILINESTRING'):
        return False
    elif geometry_type in ('POLYGON', 'MULTIPOLYGON'):
        return True
    else:
        raise WktParsingException(f'Incorrect geometry type: {geometry_type}')


def wkt_brackets(geometry_type):
    if geometry_type in ['LINESTRING']:
        return 1
    elif geometry_type in ['POLYGON', 'MULTILINESTRING']:
        return 2
    elif geometry_type in ['MULTIPOLYGON']:
        return 3
    else:
        raise WktParsingException(f'Incorrect geometry type: {geometry_type}')


def points_from_geometry(geometry):

    assert isinstance(geometry, ogr.Geometry)

    return points_from_wkt(geometry.ExportToWkt())


if __name__ == '__main__':

    def check_parsing(wkt):
        points, geometry_type = points_from_wkt(wkt)
        if points is not None:
            print(points)

    check_parsing('POINT (10 20)')
    check_parsing('CIRCULARSTRING (1 5, 6 2, 7 3)')
    check_parsing('POLYGON EMPTY')
    check_parsing('LINESTRING (10 20, 20 30)')
    check_parsing('MULTILINESTRING ((10 10, 20 20, 10 40), (40 40, 30 30, 40 20, 30 10))')
    check_parsing('MULTIPOLYGON (((40 40, 20 45, 45 30, 40 40)), ((20 35, 10 30, 10 10, 30 5, 45 20, 20 35), (30 20, 20 15, 20 25, 30 20)))')



"""
Testing problem180meridian.Geometries
"""


from osgeo import ogr

from ..problem180meridian.polygonhierarchy import PolygonHierarchy


# Testing unification 4 geometries with int coordinates,
# 1 of them is within the other one,
# and 1 more os within with 1 line touching its border
def test_geometries_int_1within_1touch():

    test_geometries = [
        ogr.Geometry(wkt='MULTIPOLYGON (((20 20, 20 21, 21 21, 21 20, 20 20)))'),
        ogr.Geometry(wkt='MULTIPOLYGON (((10 10, 10 11, 11 11, 11 10, 10 10)))'),
        ogr.Geometry(wkt='MULTIPOLYGON (((0 0, 0 1, 1 1, 1 0, 0 0)))'),
        ogr.Geometry(wkt='MULTIPOLYGON (((0 0, 0 100, 100 100, 100 0, 0 0)))'),
    ]

    hierarchy = PolygonHierarchy(*test_geometries)
    output = hierarchy.multipolygon().ExportToWkt()

    assert output == 'POLYGON ((100 0,1 0,1 1,0 1,0 100,100 100,100 0),(11 11,10 11,10 10,11 10,11 11),(21 20,21 21,20 21,20 20,21 20))'
